# $Id: UI_gtk.pm 2868 2007-07-01 07:17:14Z verthezp $
# $URL: file:///var/lib/svn/intclock/tags/R2_12/lib/Intclock/UI_gtk.pm $
#___________________________________________________________________
#
# User interface routines for intclock (Gtk interface)
#___________________________________________________________________

# Copyright (C) 2004-2007 by Peter Verthez

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.

# Written by Peter Verthez, <Peter.Verthez@advalvas.be>.
# Based on Jens-Ulrik Petersen's hsclock

# Note that Intclock::UI_gtk::conf and Intclock::UI_gtk::messagebox
# are defined at the end of this file

package Intclock::UI_gtk;

use strict;
use diagnostics;
use Gtk;
use Intclock::TZ;
use Intclock::UI;

my $true  = 1;
my $false = 0;

my $window_border_width = 3;
my $table_col_spacings  = 6;
my $table_row_spacings  = 3;
my $label_x_padding     = 3;
my $label_y_padding     = 3;

sub start {
  my ($optval, $progname, $about) = @_;
  set_locale Gtk;
  init Gtk;
  _create_ui($optval, $progname, $about);
  main Gtk;
}

sub _create_ui {
  my ($optval, $progname, $about) = @_;
  
  # main window
  my $main = new Gtk::Window("toplevel");
  $main->set_title($progname);
  $main->border_width($window_border_width);
  $main->set_policy($false, $false, $true);
  $main->signal_connect("delete_event", \&_cb_exit_ui_as_ok);

  my $menu = _create_menu($main, $optval, $progname, $about);

  # event box, to be able to capture mouse click
  my $eventbox = new Gtk::EventBox();
  $main->add($eventbox);
  $eventbox->signal_connect("event", \&_cb_button_press, $menu);

  # table for the clocks
  _create_clock_table($eventbox, $optval);

  # show all widgets
  $main->show_all();
}

sub _create_clock_table {
  my ($parent, $optval) = @_;
  # determine nr of rows and columns for table
  my ($rows, $columns);
  my $nrZones = scalar(@{$optval->{"zones"}});
  my ($layout, $layout_arg) = @{$optval->{"layout"}};
  if ($layout eq "horizontal") {
    $rows = 1; $columns = $nrZones;
  }
  elsif ($layout eq "columns") {
    use integer;
    $columns = $layout_arg;
    $columns = $nrZones if $columns > $nrZones;
    $rows    = ($nrZones - 1) / $columns + 1;
  }
  else {
    $rows = $nrZones; $columns = 1;
  }

  my $table = new Gtk::Table($rows, $columns, $true);
  $table->set_col_spacings($table_col_spacings);
  $table->set_row_spacings($table_row_spacings);
  $parent->add($table);

  _setup_clocks($table, $columns, $optval->{"zones"});
}

sub _setup_clocks {
  my ($table, $columns, $zones) = @_;
  my $t = time();
  Intclock::UI::debug($t);
  my $i = 0;
  my @clocks;
  foreach my $zone (@$zones) {
    my ($zonename) = @$zone;
    my $tz_a = Intclock::TZ::in_timezone($t, $zonename);
    my $tz_t = Intclock::UI::date_time_string($tz_a);
    my $zonetext = Intclock::UI::timezone_label($zone, $tz_a);

    my $frame = new Gtk::Frame($zonetext);
    $table->attach_defaults($frame,
			    $i % $columns, $i % $columns + 1,
			    $i / $columns, $i / $columns + 1);

    my $label = new Gtk::Label($tz_t);
    $label->set_padding($label_x_padding, $label_y_padding);
    $frame->add($label);
    $i++;
    push @clocks, [$label, $frame, $zone];
  }
  _set_timer(\@clocks);
}

sub _cb_redraw_clocks {
  my ($main, $config) = @_;
  my ($eventbox) = $main->children();
  my ($table) = $eventbox->children();
  $eventbox->remove($table);

  _create_clock_table($eventbox, $config);
  $main->show_all();
}

sub _set_timer {
  my ($clocks) = @_;
  my $timer = Intclock::UI::get_timer_interval();
  Gtk->timeout_add($timer, \&_cb_update_clocks, $clocks);
}

sub _cb_update_clocks {
  my ($clocks) = @_;
  my $t = time();
  Intclock::UI::debug($t);
  foreach my $clock (@$clocks) {
    my ($label, $frame, $zone) = @$clock;
    my ($zonename) = @$zone;
    my $tz_a = Intclock::TZ::in_timezone($t, $zonename);
    my $tz_t = Intclock::UI::date_time_string($tz_a);
    my $zonetext = Intclock::UI::timezone_label($zone, $tz_a);

    $label->set_text($tz_t);
    $frame->set_label($zonetext);
  }
  _set_timer($clocks);
  return $false;
}

sub _cb_exit_ui_as_ok {
  Gtk->main_quit();
  return $false;
}

sub _create_menu {
  my ($main, $config, $progname, $about) = @_;
  my $menu = new Gtk::Menu();
  my $menu_item;
  {
    $menu_item = new Gtk::MenuItem("Preferences");
    $menu->append($menu_item);
    $menu_item->signal_connect("activate",
			       \&Intclock::UI_gtk::conf::show,
			       $main, $config, $progname,
			       \&_cb_redraw_clocks);
    $menu_item->show();
  }
  {
    $menu_item = new Gtk::MenuItem();
    $menu->append($menu_item);
    $menu_item->show();
  }
  {
    $menu_item = new Gtk::MenuItem("About");
    $menu->append($menu_item);
    $menu_item->signal_connect("activate",
			       \&Intclock::UI_gtk::messagebox::show,
			       "About $progname", $about);
    $menu_item->show();
  }
  {
    $menu_item = new Gtk::MenuItem("Quit");
    $menu->append($menu_item);
    $menu_item->signal_connect("activate", \&_cb_exit_ui_as_ok);
    $menu_item->show();
  }
  return $menu;
}

sub _cb_button_press {
  my ($button, $menu, $event) = @_;
  return $false if not defined $event->{"type"};

  if (($event->{"type"} eq 'button_press') and ($event->{"button"} == 3)) {
    $menu->popup(undef, undef, 0, $event->{'button'}, undef);
    return $true;
  }
  
  return $false;
}

1;

###############################################################################
#  message window
###############################################################################

package Intclock::UI_gtk::messagebox;

use strict;
use diagnostics;
use Gtk;
use vars qw($aboutBox);

sub _init {
  my ($title, $contents) = @_;
  $aboutBox = new Gtk::Dialog();
  $aboutBox->set_title($title);
  # Block deletion of the window:
  $aboutBox->signal_connect("delete_event", sub { return $true; });
  
  my $frame = new Gtk::Frame();
  $aboutBox->vbox->pack_start($frame, $true, $true, 0);
  $frame->border_width(5);
  $frame->show();
  
  my $label = new Gtk::Label($contents);
  $frame->add($label);
  $label->show();
  
  my $button = new Gtk::Button("Close");
  $aboutBox->action_area->pack_start($button, $true, $true, 0);
  $button->signal_connect("clicked", \&_cb_close_dialog);
  $button->show();
}

sub show {
  my ($button, $title, $contents) = @_;
  _init($title, $contents) if not defined $aboutBox;
  $aboutBox->show();
}

sub _cb_close_dialog {
  my ($button) = @_;
  $aboutBox->hide();
}

1;

###############################################################################
#  configuration window
###############################################################################

package Intclock::UI_gtk::conf;

use strict;
use diagnostics;
use Gtk;
use POSIX qw(strftime);
use Intclock::Config;
use Intclock::TZ;
use Intclock::UI;
use vars qw($dialog $tzTree $tzList $tzCommentsEntry $tzNameEntry
	    $tzNameLabel $tzNameButton
	    $addTzButton $delTzButton $upTzButton $downTzButton $nameButton
	    $userNameButton $secondsButton $hour24Button $dateButton
	    $customFormatButton
	    $tzButton $vowelButton $horizButton $vertButton
	    $colButton $colSpin $timeFormatCombo $timeFormatLabel
	    $selectedTzInTree $selectedTzInList @specialTimezones
	    $selectedL $numInList @sample_time $initialized);

BEGIN {
  @specialTimezones = ("UTC");
  $selectedL = 0;
  $numInList = 0;
  @sample_time = localtime();
  $initialized = 0;
}

sub _init {
  my ($main, $config, $progname, $redrawfunc) = @_;
  my $spacing = 4;
  my $buttonwidth = 70;
  my $buttonheight = 30;
  my $defopts = ['fill', 'expand'];
  my $noexpand = ['fill'];
    
  $dialog = new Gtk::Dialog();
  $dialog->set_title("Preferences for $progname");
  $dialog->set_position("center");
  $dialog->set_transient_for($main);
  # Block deletion of the window:
  $dialog->signal_connect("delete_event", sub { return $true });

  # box.frame1: zone frame
  my $frame1 = new Gtk::Frame("Timezones");
  $dialog->vbox->pack_start($frame1, $true, $true, $spacing);
  $frame1->show();

  # box.frame1.table1: zone table
  my $table1 = new Gtk::Table(7, 3, $false);
  $table1->set_row_spacings(10);
  $table1->set_row_spacing(0, 0);
  $table1->set_col_spacings(10);
  $frame1->add($table1);
  $table1->show();

  # box.frame1.table1.label11
  my $label11 = new Gtk::Label("Available timezones");
  $label11->set_justify('left');
  $table1->attach($label11, 0, 1, 0, 1, $defopts, $noexpand,
		  $spacing, $spacing);
  $label11->show();

  # box.frame1.table1.label14
  my $label14 = new Gtk::Label("Selected timezones");
  $label14->set_justify('left');
  $table1->attach($label14, 2, 3, 0, 1, $defopts, $noexpand,
		  $spacing, $spacing);
  $label14->show();

  # box.frame1.table1.scrolled11: scrolled window for timezone tree
  my $scrolled11 = new Gtk::ScrolledWindow(undef, undef);
  $scrolled11->set_usize(300, 200);
  $table1->attach($scrolled11, 0, 1, 1, 6, $defopts, $defopts,
		  $spacing, $spacing);
  $scrolled11->set_policy('automatic', 'automatic');
  $scrolled11->show();

  # box.frame1.table1.scrolled11.tzTree: timezone tree
  $tzTree = new Gtk::Tree();
  $scrolled11->add_with_viewport($tzTree);
  $tzTree->set_selection_mode('single');
  $tzTree->set_view_mode('item');
  $tzTree->show();

  # add, remove, up and down buttons
  $addTzButton = new Gtk::Button("Add");
  $addTzButton->set_usize($buttonwidth, $buttonheight);
  $addTzButton->set_sensitive($false);
  $addTzButton->signal_connect("clicked", \&_cb_add_timezone,
			       [ $main, $config, $redrawfunc ]);
  $table1->attach($addTzButton, 1, 2, 1, 2, [], [], $spacing, $spacing);
  $addTzButton->show();
  
  $delTzButton = new Gtk::Button("Remove");
  $delTzButton->set_usize($buttonwidth, $buttonheight);
  $delTzButton->set_sensitive($false);
  $delTzButton->signal_connect("clicked", \&_cb_remove_timezone,
			       [ $main, $config, $redrawfunc ]);
  $table1->attach($delTzButton, 1, 2, 2, 3, [], [], $spacing, $spacing);
  $delTzButton->show();
  
  $upTzButton = new Gtk::Button("Up");
  $upTzButton->set_usize($buttonwidth, $buttonheight);
  $upTzButton->set_sensitive($false);
  $upTzButton->signal_connect("clicked", \&_cb_move_timezone_up,
			       [ $main, $config, $redrawfunc ]);
  $table1->attach($upTzButton, 1, 2, 3, 4, [], [], $spacing, $spacing);
  $upTzButton->show();
  
  $downTzButton = new Gtk::Button("Down");
  $downTzButton->set_usize($buttonwidth, $buttonheight);
  $downTzButton->set_sensitive($false);
  $downTzButton->signal_connect("clicked", \&_cb_move_timezone_down,
			       [ $main, $config, $redrawfunc ]);
  $table1->attach($downTzButton, 1, 2, 4, 5, [], [], $spacing, $spacing);
  $downTzButton->show();

  # scrolled window for timezone selection
  my $scrolled14 = new Gtk::ScrolledWindow(undef, undef);
  $scrolled14->set_usize(200, 200);
  $table1->attach($scrolled14, 2, 3, 1, 6, $defopts, $defopts,
		  $spacing, $spacing);
  $scrolled14->set_policy('automatic', 'automatic');
  $scrolled14->show();

  # timezone selection
  $tzList = new Gtk::CList(1);
  $scrolled14->add_with_viewport($tzList);
  $tzList->set_selection_mode('single');
  $tzList->set_shadow_type('none');
  $tzList->signal_connect("select_row", \&_cb_select_zone_in_list);
  $tzList->signal_connect("unselect_row", \&_cb_deselect_zone_in_list);
  $tzList->show();

  # comments field
  $tzCommentsEntry = new Gtk::Entry();
  $table1->attach($tzCommentsEntry, 0, 1, 6, 7, $defopts, $defopts,
		  $spacing, $spacing);
  $tzCommentsEntry->show();

  # box for timezone name
  my $box16 = new Gtk::HBox($false, $spacing);
  $table1->attach($box16, 2, 3, 6, 7, $defopts, $defopts,
		  $spacing, $spacing);
  $box16->show();

  # timezone name label
  $tzNameLabel = new Gtk::Label("User Defined name");
  $tzNameLabel->set_justify('left');
  $tzNameLabel->set_sensitive($false);
  $box16->pack_start($tzNameLabel, $false, $true, $spacing);
  $tzNameLabel->show(); 

  # timezone name field
  $tzNameEntry = new Gtk::Entry();
  $box16->pack_start($tzNameEntry, $false, $true, $spacing);
  $tzNameEntry->set_sensitive($false);
  $tzNameEntry->show();

  # timezone name button
  $tzNameButton = new Gtk::Button("Modify");
  $tzNameButton->set_usize($buttonwidth, $buttonheight);
  $tzNameButton->set_sensitive($false);
  $tzNameButton->signal_connect("clicked", \&_cb_change_timezone_user_name,
				[ $main, $config, $redrawfunc ]);
  $box16->pack_start($tzNameButton, $false, $true, $spacing);
  $tzNameButton->show();
  
  # horizontal box for options
  my $box2 = new Gtk::HBox($false, $spacing);
  $dialog->vbox->pack_start($box2, $false, $true, $spacing);
  $box2->show();

  # display frame
  my $frame21 = new Gtk::Frame("Display");
  $box2->pack_start($frame21, $true, $true, $spacing);
  $frame21->show();

  # checkboxes
  my $table21 = new Gtk::Table(3, 2, $false);
  $frame21->add($table21);
  $tzButton = new Gtk::RadioButton("Timezone abbreviation");
  $tzButton->signal_connect("toggled", \&_cb_toggle_display_option,
			    [ $main, $config, $redrawfunc ]);
  $table21->attach_defaults($tzButton, 0, 2, 0, 1);
  $tzButton->show();
  $nameButton = new Gtk::RadioButton("City name", $tzButton);
  $nameButton->signal_connect("toggled", \&_cb_toggle_display_option,
			      [ $main, $config, $redrawfunc ]);
  $table21->attach_defaults($nameButton, 0, 2, 1, 2);
  $nameButton->show();
  $vowelButton = new Gtk::CheckButton("Drop vowels");
  $vowelButton->signal_connect("toggled", \&_cb_apply_change,
			       [ $main, $config, $redrawfunc ]);
  $table21->attach_defaults($vowelButton, 1, 2, 2, 3);
  $vowelButton->set_sensitive($false);
  $vowelButton->show();
  $userNameButton = new Gtk::RadioButton("User Defined name", $tzButton);
  $userNameButton->signal_connect("toggled", \&_cb_toggle_display_option,
				  [ $main, $config, $redrawfunc ]);
  $table21->attach_defaults($userNameButton, 0, 2, 3, 4);
  $userNameButton->show();
  $table21->show();
  
  # orientation frame
  my $frame22 = new Gtk::Frame("Orientation");
  $box2->pack_start($frame22, $true, $true, $spacing);
  $frame22->show();

  # radiobuttons
  my $box22 = new Gtk::VBox($false, $spacing);
  $frame22->add($box22);
  
  $vertButton = new Gtk::RadioButton("Vertical (1 column)");
  $vertButton->signal_connect("toggled", \&_cb_apply_change,
			      [ $main, $config, $redrawfunc ]);
  $box22->pack_start($vertButton, $true, $true, 0);
  $vertButton->show();
  
  $horizButton = new Gtk::RadioButton("Horizontal (1 row)", $vertButton);
  $horizButton->signal_connect("toggled", \&_cb_apply_change,
			       [ $main, $config, $redrawfunc ]);
  $box22->pack_start($horizButton, $true, $true, 0);
  $horizButton->show();
  
  my $box221 = new Gtk::HBox($false, $spacing);
  $box22->pack_start($box221, $true, $true, 0);
  $box221->show();

  $colButton = new Gtk::RadioButton("Number of columns", $vertButton);
  $colButton->signal_connect("toggled", \&_cb_toggle_column_entry,
				  [ $main, $config, $redrawfunc ]);
  $box221->pack_start($colButton, $true, $true, 0);
  $colButton->show();
  
  my $adj = new Gtk::Adjustment(2, 2, 10, 1, 1, 1);
  $colSpin = new Gtk::SpinButton($adj, 0, 0);
  $adj->signal_connect("value_changed", \&_cb_apply_change,
		       [ $main, $config, $redrawfunc ]);
  $box221->pack_start($colSpin, $true, $true, 0);
  $colSpin->show();
  $colSpin->set_sensitive($false);
  
  $box22->show();

  # horizontal box for time format and sample output
  my $box3 = new Gtk::HBox($false, $spacing);
  $dialog->vbox->pack_start($box3, $false, $true, $spacing);
  $box3->show();

  # frame for time format selection
  my $frame31 = new Gtk::Frame("Format of time display");
  $box3->pack_start($frame31, $true, $true, $spacing);
  $frame31->show();

  # vertical packing for widgets
  my $box31 = new Gtk::VBox($false, $spacing);
  $frame31->add($box31);
  $box31->show();

  # buttons for time format
  $secondsButton = new Gtk::CheckButton("Show seconds");
  $box31->pack_start($secondsButton, $true, $true, 0);
  $secondsButton->signal_connect("toggled", \&_cb_set_time_format,
				 [ $main, $config, $redrawfunc ]);
  $secondsButton->show();
  
  $hour24Button = new Gtk::CheckButton("Show 24 hour format");
  $box31->pack_start($hour24Button, $true, $true, 0);
  $hour24Button->signal_connect("toggled", \&_cb_set_time_format,
				[ $main, $config, $redrawfunc ]);
  $hour24Button->show();
  
  $dateButton = new Gtk::CheckButton("Show date");
  $box31->pack_start($dateButton, $true, $true, 0);
  $dateButton->signal_connect("toggled", \&_cb_set_time_format,
			      [ $main, $config, $redrawfunc ]);
  $dateButton->show();

  # horizontal packing for custom format and sample
  my $box311 = new Gtk::HBox($false, $spacing);
  $box31->pack_start($box311, $true, $true, 0);
  $box311->show();

  # button for custom format
  $customFormatButton = new Gtk::CheckButton("Custom format");
  $box311->pack_start($customFormatButton, $true, $true, 0);
  $customFormatButton->signal_connect("toggled", \&_cb_set_time_format,
				    [ $main, $config, $redrawfunc, "custom" ]);
  $customFormatButton->show();

  # combobox for selecting time format
  $timeFormatCombo = new Gtk::Combo();
  $timeFormatCombo->set_popdown_strings(@Intclock::UI::timeformats);
  $timeFormatCombo->entry->signal_connect("changed",
					  \&_cb_apply_change,
					  [ $main, $config, $redrawfunc ]);
  $box311->pack_start($timeFormatCombo, $true, $true, $spacing);
  $timeFormatCombo->show();

  # Close button
  my $closeButton = new Gtk::Button("Close");
  $dialog->action_area->pack_start($closeButton, $true, $true, 0);
  $closeButton->signal_connect("clicked", sub { $dialog->hide() });
  $closeButton->show();
  
  _fill_timezone_tree();
}

sub show {
  my ($button, $main, $config, $progname, $redrawfunc) = @_;
  $initialized = 0;
  _init($main, $config, $progname, $redrawfunc) if not defined $dialog;
  _ui_load_config($config);
  $dialog->show();
  $initialized = 1;
}

sub _apply {
  my ($main, $config, $redrawfunc) = @_;
  if ($initialized and _ui_save_config($config)) {
    Intclock::UI::set_config($config);
    &{$redrawfunc}($main, $config);
  }
}

sub _add_timezone_tree_branch {
  my ($tree, $label) = @_;
  
  my $treeitem = new_with_label Gtk::TreeItem($label);
  $treeitem->show();
  $tree->append($treeitem);

  my $subtree = new Gtk::Tree();
  $treeitem->set_subtree($subtree);
  $treeitem->signal_connect('collapse', \&_cb_collapse_subtree);
  $treeitem->signal_connect('select', \&_cb_select_zone_in_tree, undef);
  $treeitem->signal_connect('deselect', \&_cb_deselect_zone_in_tree, undef);

  return $subtree;
}

sub _add_timezone_tree_leaf {
  my ($tree, $label, $selectInfo) = @_;

  my $treeitem = new_with_label Gtk::TreeItem($label);
  $treeitem->show();
  $treeitem->signal_connect('select', \&_cb_select_zone_in_tree, $selectInfo);
  $treeitem->signal_connect('deselect', \&_cb_deselect_zone_in_tree, $selectInfo);
  $tree->append($treeitem);
}

sub _tz_text {
  my (@tzAbbr) = @_;
  if ($tzAbbr[0] eq $tzAbbr[1]) {
    return $tzAbbr[0];
  }
  else {
    return "$tzAbbr[0]/$tzAbbr[1]";
  }
}

sub _zones_by_longitude {
  my ($zone_l) = @_;
  my @keys = keys %$zone_l;
  @keys = sort { $zone_l->{$a}->{"long"} <=> $zone_l->{$b}->{"long"} } @keys;
  return @keys;
}

sub _fill_timezone_tree {
  my ($treeitem, $subtree, $subsubtree);
  for my $timezone (@specialTimezones) {
    _add_timezone_tree_leaf($tzTree, $timezone, [undef, undef, $timezone]);
  }
  my $zonehash = Intclock::TZ::zone_info();
  foreach my $cont (sort keys %$zonehash) {
    my $subtree = _add_timezone_tree_branch($tzTree, $cont);
    my $zone_c = $zonehash->{$cont};
    foreach my $country (sort keys %$zone_c) {
      my $zone_l = $zone_c->{$country};
      if (scalar keys %$zone_l == 1) {
	my ($zone) = keys %$zone_l;
	my $tzAbbr = _tz_text(Intclock::TZ::get_timezone_names("$cont/$zone"));
	my $tzone = $zone;
	$tzone =~ s/_/ /g;
	_add_timezone_tree_leaf($subtree, "$country - $tzone [$tzAbbr]",
				[$cont, $country, $zone]);
      }
      else {
	$subsubtree = _add_timezone_tree_branch($subtree, $country);
	foreach my $zone (_zones_by_longitude $zone_l) {
	  my $tzAbbr = _tz_text(Intclock::TZ::get_timezone_names("$cont/$zone"));
	  my $tzone = $zone;
	  $tzone =~ s/_/ /g;
	  _add_timezone_tree_leaf($subsubtree, "$tzone [$tzAbbr]",
				  [$cont, $country, $zone]);
	}
      }
    }
  }
}

sub _cb_toggle_display_option {
  my ($widget, $data) = @_;
  my ($main, $config, $redrawfunc) = @$data;
  $vowelButton->set_sensitive($widget->active());
  $vowelButton->set_active($false);
  _apply($main, $config, $redrawfunc);
}

sub _cb_toggle_column_entry {
  my ($widget, $data) = @_;
  my ($main, $config, $redrawfunc) = @$data;
  $colSpin->set_sensitive($widget->active());
  _apply($main, $config, $redrawfunc);
}

sub _cb_apply_change {
  my ($widget, $data) = @_;
  my ($main, $config, $redrawfunc) = @$data;
  _apply($main, $config, $redrawfunc);
}

sub _cb_select_zone_in_tree {
  my ($widget, $path) = @_;
  if ($path) {
    my ($cont, $country, $zone) = @$path;
    if (defined $cont) {
      my $zonehash = Intclock::TZ::zone_info();
      my $zoneinfo = $zonehash->{$cont}->{$country}->{$zone};
      my $continent = $zoneinfo->{"continent"};
      $selectedTzInTree = "$continent/$zone";
      my $comment = $zoneinfo->{"comments"};
      $comment ||= $country;
      $tzCommentsEntry->set_text($comment);
    }
    else {
      $selectedTzInTree = $zone;
    }
    $addTzButton->set_sensitive($true);
  }
}

sub _cb_deselect_zone_in_tree {
  my ($widget, $path) = @_;
  if ($path) {
    undef $selectedTzInTree;
    $tzCommentsEntry->set_text("");
    $addTzButton->set_sensitive($false);
  }
}

sub _cb_collapse_subtree {
  my ($widget) = @_;
  my $tree = $widget->subtree;
  my ($selected) = $tzTree->selection();
  if ($selected) {
    my $item = $selected;
    while ($item->parent) {
      $item = $item->parent();
      if ($item == $tree) {
	$selected->deselect();
	last;
      }
    }
  }
}

sub _get_strings_from_tzlist {
  my ($label) = @_;
  my @splittext = split '"', $label;
  $splittext[0] =~ s/ $//;
  return @splittext;
}

sub _make_string_for_tzlist {
  my ($zonename, $username) = @_;
  if ($username) {
    return "$zonename \"$username\"";
  }
  else {
    return $zonename;
  }
}

sub _set_tz_name_sensitivity {
  $tzNameLabel->set_sensitive($selectedL == 1 ? $true : $false);
  $tzNameEntry->set_sensitive($selectedL == 1 ? $true : $false);
  $tzNameButton->set_sensitive($selectedL == 1 ? $true : $false);
  if ($selectedL == 1) {
    my @splittext
	= _get_strings_from_tzlist($tzList->get_text($selectedTzInList, 0));
    $tzNameEntry->set_text(defined $splittext[1] ? $splittext[1] : "");
  }
  else {
    $tzNameEntry->set_text("");
  }
}

sub _cb_select_zone_in_list {
  my ($widget, $row) = @_;
  $selectedTzInList = $row;
  $selectedL++;
  $delTzButton->set_sensitive($selectedL and ($numInList > 1) ? $true : $false);
  $upTzButton->set_sensitive($row > 0 ? $true : $false);
  $downTzButton->set_sensitive($row < $numInList - 1 ? $true : $false);
  _set_tz_name_sensitivity();
}

sub _cb_deselect_zone_in_list {
  my ($widget, $row) = @_;
  undef $selectedTzInList;
  $selectedL--;
  $delTzButton->set_sensitive($selectedL ? $true : $false);
  if (! $selectedL) {
    $upTzButton->set_sensitive($false);
    $downTzButton->set_sensitive($false);
  }
  _set_tz_name_sensitivity();
}

sub _add_timezone_helper {
  my ($tz) = @_;
  $numInList++;
  $tzList->append(_make_string_for_tzlist(@$tz));
}

sub _cb_add_timezone {
  my ($widget, $data) = @_;
  my ($main, $config, $redrawfunc) = @$data;
  _add_timezone_helper([$selectedTzInTree, ""]);
  _apply($main, $config, $redrawfunc);
}

sub _cb_remove_timezone {
  my ($widget, $data) = @_;
  my ($main, $config, $redrawfunc) = @$data;
  my $row = $selectedTzInList;
  $numInList--;
  $tzList->remove($selectedTzInList);
  _apply($main, $config, $redrawfunc);
  if ($row < $numInList) {
    $tzList->select_row($row, 0);
  }
}

sub _cb_move_timezone_up {
  my ($widget, $data) = @_;
  my ($main, $config, $redrawfunc) = @$data;
  my $row = $selectedTzInList;
  $tzList->unselect_row($row, 0);
  $tzList->row_move($row, $row - 1);
  $tzList->select_row($row - 1, 0);
  _apply($main, $config, $redrawfunc);
}

sub _cb_move_timezone_down {
  my ($widget, $data) = @_;
  my ($main, $config, $redrawfunc) = @$data;
  my $row = $selectedTzInList;
  my $text = $tzList->get_text($row, 0);
  $tzList->unselect_row($row, 0);
  $tzList->row_move($row, $row + 1);
  $tzList->select_row($row + 1, 0);
  _apply($main, $config, $redrawfunc);
}

sub _cb_change_timezone_user_name {
  my ($widget, $data) = @_;
  my ($main, $config, $redrawfunc) = @$data;
  my $row = $selectedTzInList;
  my @oldtext = _get_strings_from_tzlist($tzList->get_text($row, 0));
  my $input = Intclock::Config::fix_utf8($tzNameEntry->get_text());
  $tzList->set_text($row, 0, _make_string_for_tzlist($oldtext[0], $input));
  _apply($main, $config, $redrawfunc);
}

sub _calculate_time_format_buttons {
  my ($initial) = @_;
  my $format = $timeFormatCombo->entry->get_text();
  my $predefined = exists $Intclock::UI::tf_configs{$format};

  if ($initial) {
    $customFormatButton->set_active($predefined ? $false : $true);
    $timeFormatCombo->set_sensitive($predefined ? $false : $true);
  }
  
  if ($predefined) {
    my $config = $Intclock::UI::tf_configs{$format};
    $secondsButton->set_active($config->[0] ? $true : $false);
    $hour24Button->set_active($config->[1] ? $true : $false);
    $dateButton->set_active($config->[2] ? $true : $false);
  }
  else {
    $secondsButton->set_sensitive($false);
    $hour24Button->set_sensitive($false);
    $dateButton->set_sensitive($false);
  }
}

sub _recalculate_time_format {
  my $timeformat = "";
  my $seconds = "";
  if ($dateButton->get_active()) {
    $timeformat .= "%a %d%n";
  }
  if ($secondsButton->get_active()) {
    $seconds = ":%S";
  }
  if ($hour24Button->get_active()) {
    $timeformat .= "%H:%M$seconds";
  }
  else {
    $timeformat .= "%I:%M$seconds %p";
  }
  $timeFormatCombo->entry->set_text($timeformat);
}

sub _cb_set_time_format {
  my ($widget, $data) = @_;
  my ($main, $config, $redrawfunc, $custom) = @$data;
  if ($custom and $custom eq "custom") {
    my $state = $customFormatButton->get_active();
    $timeFormatCombo->set_sensitive($state);
    $secondsButton->set_sensitive($state ? $false : $true);
    $hour24Button->set_sensitive($state ? $false : $true);
    $dateButton->set_sensitive($state ? $false : $true);
    if (not $state) {
      _calculate_time_format_buttons($false);
    }
  }
  else {
    _recalculate_time_format();
  }
  _apply($main, $config, $redrawfunc);
}

sub _ui_load_config {
  my ($config) = @_;
  $tzList->clear();
  if ($config->{"zones"}) {
    foreach my $timezone (@{$config->{"zones"}}) {
      _add_timezone_helper($timezone);
    }
  }
  if ($config->{"name"}) {
    if ($config->{"name"} == 1) {
      $nameButton->set_active($true);
    }
    elsif ($config->{"name"} == 2) {
      $userNameButton->set_active($true);
    }
  }
  if ($config->{"vowel"}) {
    $vowelButton->set_active($true);
  }
  my ($layout, $layout_arg) = @{$config->{"layout"}};
  if ($layout eq "columns") {
    $colButton->set_active($true);
    $colSpin->set_sensitive($true);
    $colSpin->set_value($layout_arg);
  }
  elsif ($layout eq "horizontal") {
    $horizButton->set_active($true);
  }
  elsif ($layout eq "vertical") {
    $vertButton->set_active($true);
  }
  if ($config->{"format"}) {
    my $tf = $config->{"format"};
    $timeFormatCombo->entry->set_text($tf);
    _calculate_time_format_buttons($true);
  }
  # Needed because $tzList is cleared
  $selectedL = 0;
  _set_tz_name_sensitivity();
}

sub _ui_save_config {
  my ($config) = @_;
  my ($options, @timezones) = "";
  for (my $i = 0; $i < $tzList->rows(); $i++) {
    my @splittext = _get_strings_from_tzlist($tzList->get_text($i, 0));
    push @timezones, [ $splittext[0], $splittext[1] || "" ];
  }
  return undef if ! @timezones;
  $config->{"zones"}      = \@timezones;
  $config->{"name"}
  = ($userNameButton->get_active() ?
     2 : ($nameButton ->get_active() ?
	  1 :
	  0));
  $config->{"vowel"}      = ($vowelButton->get_active() ? 1 : 0);
  if ($colButton->get_active()) {
    $config->{"layout"} = [ "columns", $colSpin->get_value_as_int() ];
  }
  elsif ($horizButton->get_active()) {
    $config->{"layout"} = [ "horizontal" ];
  }
  elsif ($vertButton->get_active()) {
    $config->{"layout"} = [ "vertical" ];
  }
  $config->{"format"}     = $timeFormatCombo->entry->get_text();

  Intclock::Config::save($config);
  return 1;
}

1;
